@tool
extends ShaderMaterial
class_name StylizedParticle3DMaterial

const SHADER_PROP_PREFIX = "SH_"

var shader_configuration: ShaderConfigurationFile

@export
var flipbook_resource: FlipbookResource:
	get:
		return flipbook_resource
	set(val):
		if flipbook_resource:
			flipbook_resource.changed.disconnect(Callable(self, "_update_flipbook_values"))
		if val:
			val.changed.connect(Callable(self, "_update_flipbook_values"))
		flipbook_resource = val
		_update_flipbook_values()

func _update_flipbook_values():
	print('updating flipbook')
	if flipbook_resource != null:
		shader_configuration.sh_flipbook_enabled = true
		set_shader_parameter("flipbook", flipbook_resource.flipbook)
		set_shader_parameter("particles_anim_h_frames", flipbook_resource.h_frames)
		set_shader_parameter("particles_anim_v_frames", flipbook_resource.v_frames)
	else:
		shader_configuration.sh_flipbook_enabled = false
		set_shader_parameter("flipbook", null)

func _init():
	print("init shader")
	shader_configuration = load("res://addons/vfx_kit/StylizedParticle3DShaderProxy.gd").new()
	shader_configuration.changed.connect(Callable(self, "emit_changed"))
	shader = shader_configuration.shader
	# TODO
	# don't make a unique shader ?


func _get_property_list():
	var extra_prop = []
	if shader_configuration!= null:
		for prop in shader_configuration.get_property_list():
			if prop.name.begins_with("sh_"):
				prop.name = SHADER_PROP_PREFIX + prop.name.trim_prefix("sh_")
				extra_prop.append(prop)
	return extra_prop

func _set(property, value):
	if shader_configuration!= null:
		if property.begins_with(SHADER_PROP_PREFIX):
			shader_configuration.set("sh_" + property.trim_prefix(SHADER_PROP_PREFIX), value)
			emit_changed()
			return true
	return false

func _get(property):
	if shader_configuration!= null:
		if property.begins_with(SHADER_PROP_PREFIX):
			return shader_configuration.get("sh_" + property.trim_prefix(SHADER_PROP_PREFIX))
	else:
		return null
