@tool
extends Resource
class_name ShaderConfigurationFile

var shader: Shader

enum BlendMode{
	ADD,
	MIX
}

enum BillboardMode{
	NO_BILLBOARD,
	ANGLE_FROM_PARTICLE,
	LOCAL_Y_AXIS,
	PROJECTED_LOCAL_Y
}

var billboard_codes = {
	BillboardMode.NO_BILLBOARD : "",
	BillboardMode.ANGLE_FROM_PARTICLE : """
	mat4 mat_world = mat4(normalize(INV_VIEW_MATRIX[0]), normalize(INV_VIEW_MATRIX[1]) ,normalize(INV_VIEW_MATRIX[2]), MODEL_MATRIX[3]);
	mat_world = mat_world * mat4(vec4(cos(custom_custom.x), -sin(custom_custom.x), 0.0, 0.0), vec4(sin(custom_custom.x), cos(custom_custom.x), 0.0, 0.0), vec4(0.0, 0.0, 1.0, 0.0), vec4(0.0, 0.0, 0.0, 1.0));
	MODELVIEW_MATRIX = VIEW_MATRIX * mat_world;
	MODELVIEW_MATRIX = MODELVIEW_MATRIX * mat4(vec4(length(MODEL_MATRIX[0].xyz), 0.0, 0.0, 0.0),vec4(0.0, length(MODEL_MATRIX[1].xyz), 0.0, 0.0), vec4(0.0, 0.0, length(MODEL_MATRIX[2].xyz), 0.0), vec4(0.0, 0.0, 0.0, 1.0));
	MODELVIEW_NORMAL_MATRIX = mat3(MODELVIEW_MATRIX);
	""",
	BillboardMode.LOCAL_Y_AXIS: """
	// We will make the new model matrix with 
	vec3 new_x_axis = cross(normalize(MODEL_MATRIX[1].xyz), inverse(VIEW_MATRIX)[2].xyz);
	vec3 new_z_axis = cross(new_x_axis, normalize(MODEL_MATRIX[1].xyz));
	mat3 rot = mat3(new_x_axis, normalize(MODEL_MATRIX[1].xyz), new_z_axis);
	mat4 new_model = mat4(rot);
	new_model[3] = MODEL_MATRIX[3];
	MODELVIEW_MATRIX = VIEW_MATRIX * new_model;
	MODELVIEW_MATRIX = MODELVIEW_MATRIX * mat4(vec4(length(MODEL_MATRIX[0].xyz), 0.0, 0.0, 0.0),vec4(0.0, length(MODEL_MATRIX[1].xyz), 0.0, 0.0), vec4(0.0, 0.0, length(MODEL_MATRIX[2].xyz), 0.0), vec4(0.0, 0.0, 0.0, 1.0));
	MODELVIEW_NORMAL_MATRIX = mat3(MODELVIEW_MATRIX);
	""",
	BillboardMode.PROJECTED_LOCAL_Y: """
	vec2 camera_space_y = (VIEW_MATRIX * MODEL_MATRIX[1]).xy;
	camera_space_y = normalize(camera_space_y);
	float angle = atan(camera_space_y.x, camera_space_y.y);
	mat4 mat_world = mat4(normalize(INV_VIEW_MATRIX[0]), normalize(INV_VIEW_MATRIX[1]) ,normalize(INV_VIEW_MATRIX[2]), MODEL_MATRIX[3]);
	mat_world = mat_world * mat4(vec4(cos(angle), -sin(angle), 0.0, 0.0), vec4(sin(angle), cos(angle), 0.0, 0.0), vec4(0.0, 0.0, 1.0, 0.0), vec4(0.0, 0.0, 0.0, 1.0));
	MODELVIEW_MATRIX = VIEW_MATRIX * mat_world;
	MODELVIEW_MATRIX = MODELVIEW_MATRIX * mat4(vec4(length(MODEL_MATRIX[0].xyz), 0.0, 0.0, 0.0),vec4(0.0, length(MODEL_MATRIX[1].xyz), 0.0, 0.0), vec4(0.0, 0.0, length(MODEL_MATRIX[2].xyz), 0.0), vec4(0.0, 0.0, 0.0, 1.0));
	MODELVIEW_NORMAL_MATRIX = mat3(MODELVIEW_MATRIX);
	"""
}

var flipbook_code = {
	"uniforms": """
	uniform int particles_anim_h_frames = 1;
	uniform int particles_anim_v_frames = 1;
	uniform bool particles_anim_loop = false;
	""",
	"code": """
	float h_frames = float(particles_anim_h_frames);
	float v_frames = float(particles_anim_v_frames);
	float particle_total_frames = float(particles_anim_h_frames * particles_anim_v_frames);
	float particle_frame = floor(custom_custom.z * float(particle_total_frames));
	if (!particles_anim_loop) {
		particle_frame = clamp(particle_frame, 0.0, particle_total_frames - 1.0);
	} else {
		particle_frame = mod(particle_frame, particle_total_frames);
	}
	UV /= vec2(h_frames, v_frames);
	UV += vec2(mod(particle_frame, h_frames) / h_frames, floor((particle_frame + 0.5) / h_frames) / v_frames);
	"""
}

enum ColorMode {
	GRADIENT_MAP,
	GRADIENT_MAP_ALPHA_MULTIPLIED,
	SOURCE_FLIPBOOK,
}

var color_mode_uniforms = {
	ColorMode.SOURCE_FLIPBOOK: "uniform sampler2D flipbook : source_color, filter_linear;",
	ColorMode.GRADIENT_MAP: """
	uniform sampler2D flipbook: source_color, filter_linear;
	uniform sampler2D color_gradient: source_color, filter_linear, repeat_disable;
	""",
	ColorMode.GRADIENT_MAP_ALPHA_MULTIPLIED: """
	uniform sampler2D flipbook: source_color, filter_linear;
	uniform sampler2D color_gradient: source_color, filter_linear, repeat_disable;
	""",
}

var color_mode_code = {
	ColorMode.SOURCE_FLIPBOOK : """
	ALBEDO = color.rgb * COLOR.rgb;
	""",
	ColorMode.GRADIENT_MAP: """
	color = texture(color_gradient, vec2(color.r + color_offset));
	ALBEDO = color.rgb * COLOR.rgb;
	""",
	ColorMode.GRADIENT_MAP_ALPHA_MULTIPLIED:"""
	color = texture(color_gradient, vec2(color.r * alpha + color_offset));
	ALBEDO = color.rgb * COLOR.rgb;
	""",
}


enum AlphaMode {
	SOURCE_OPACITY,
	DISSOLVE
}

var alpha_uniforms = {
	AlphaMode.SOURCE_OPACITY: "",
	AlphaMode.DISSOLVE: """
	uniform sampler2D dissolve_texture_detail: hint_default_white;\n
	uniform sampler2D dissolve_texture_overlay_direction: hint_default_white;
	uniform float dissolve_hardness;
	instance uniform vec2 dissolve_scale = vec2(1.0);
	float overlay(float base, float blend){
	float limit = step(0.5, base);
	return mix(2.0 * base * blend, 1.0 - 2.0 * (1.0 - base) * (1.0 - blend), limit);
	}

	"""
}

var alpha_code = {
	AlphaMode.SOURCE_OPACITY: """
	alpha = COLOR.a * color.a;
	ALPHA = COLOR.a * color.a;
	""",
	AlphaMode.DISSOLVE: """
	float dissolve_main = texture(dissolve_texture_detail, full_range_uv * dissolve_scale).r;
	float dissolve_detail = texture(dissolve_texture_overlay_direction, full_range_uv * dissolve_scale).r;\n
	float erosion = (1.0 - COLOR.a) * (1.0 + dissolve_hardness);
	float dissolve_value = dissolve_main < 0.5 ? (2.0 * dissolve_main * dissolve_detail) : (1.0 - 2.0 * (1.0 - dissolve_main) * (1.0 - dissolve_detail));
	alpha = smoothstep(erosion- dissolve_hardness, erosion, dissolve_value);\n
	ALPHA = smoothstep(erosion - dissolve_hardness, erosion, dissolve_value);\n
	ALPHA *= color.a;\n
	"""
}

enum UvManipulation{
	NONE,
	SCALE_OFFSET,
	BEGIN_END_CURVE
}

var uv_manip_uniforms = {
	UvManipulation.NONE: "",
	UvManipulation.SCALE_OFFSET: """
	instance uniform vec2 uv_scale = vec2(1.0);\n
	instance uniform vec2 uv_offset;\n
	varying vec2 alpha_clamp_range;
	uniform bool clamp_alpha_x = false;
	uniform bool clamp_alpha_y = false;
	""",
	UvManipulation.BEGIN_END_CURVE: """
	uniform sampler2D offsets_curve: repeat_disable;
	varying vec2 alpha_clamp_range;
	uniform bool clamp_alpha_x = false;
	uniform bool clamp_alpha_y = false;
	"""
}

var uv_manip_code = {
	UvManipulation.NONE: "",
	UvManipulation.SCALE_OFFSET:"""
	UV = UV * uv_scale + uv_offset;
	alpha_clamp_range = vec2(0.0, 1.0);\n
	""",
	UvManipulation.BEGIN_END_CURVE:"""
	vec2 offset_scale = texture(offsets_curve, vec2(custom_custom.y/custom_custom.w)).rg;
	UV = vec2((UV.x - offset_scale.x)/(offset_scale.y - offset_scale.x), UV.y);
	alpha_clamp_range = vec2(0.0,1.0);
	"""
	
}

var color_distortion_uniform = """
instance uniform vec2 color_distortion_direction = vec2(0.0,1.0);
instance uniform float color_distortion_pan_speed = 0.1;
instance uniform float color_distortion_affect: hint_range (-1.0, 1.0, 0.001) = 0.2;
instance uniform vec2 color_distortion_scale = vec2(1.0);
uniform sampler2D color_distortion_noise;
"""

var color_distortion_code = """
	color_offset = texture(color_distortion_noise, full_range_uv * color_distortion_scale + normalize(color_distortion_direction) * TIME * color_distortion_pan_speed).r * color_distortion_affect;
"""

#TODO
enum TextureMode{
	FLIPBOOK,
	MASKED_FLIPBOOK,
	MASKED_TEXTURE,
	DOUBLE_MASKED_TEXTURE
}

@export
var sh_uv_manip_mode: UvManipulation = UvManipulation.NONE:
	get:
		return sh_uv_manip_mode
	set(val):
		sh_uv_manip_mode = val
		_generate_code()

@export
var sh_color_mode: ColorMode = ColorMode.GRADIENT_MAP:
	get:
		return sh_color_mode
	set(val):
		sh_color_mode = val
		_generate_code()

@export
var sh_alpha_mode: AlphaMode = AlphaMode.SOURCE_OPACITY:
	get:
		return sh_alpha_mode
	set(val):
		sh_alpha_mode = val
		_generate_code()

@export
var sh_blend_mode: BlendMode = BlendMode.MIX :
	get:
		return sh_blend_mode
	set(val):
		sh_blend_mode = val
		_generate_code()

@export
var sh_billboard_mode: BillboardMode:
	get:
		return sh_billboard_mode
	set(val):
		sh_billboard_mode = val
		_generate_code()

@export
var sh_flipbook_enabled = true:
	get:
		return sh_flipbook_enabled
	set(val):
		sh_flipbook_enabled = val
		_generate_code()

@export
var sh_expose_custom = false:
	get:
		return sh_expose_custom
	set(val):
		sh_expose_custom = val
		_generate_code()

@export
var sh_uv_swap = false:
	get:
		return sh_uv_swap
	set(val):
		sh_uv_swap = val
		_generate_code()

@export
var sh_uv_dissolve_full_quad= false:
	get:
		return sh_uv_dissolve_full_quad
	set(val):
		sh_uv_dissolve_full_quad = val
		_generate_code()

@export
var sh_apply_color_distortion= false:
	get:
		return sh_apply_color_distortion
	set(val):
		sh_apply_color_distortion = val
		_generate_code()

func _init():
	shader = Shader.new()
	shader.changed.connect(Callable(self, "emit_changed"))
	_generate_code()
	
func _generate_code():
	var tmp_code = ""
	tmp_code = ""
	tmp_code += "shader_type spatial;\nrender_mode "
	match sh_blend_mode:
		BlendMode.ADD:
			tmp_code += "blend_add"
		BlendMode.MIX:
			tmp_code += "blend_mix"
		_:
			print("shouldn't happen, report bug please.")
	tmp_code += ", unshaded,"
	tmp_code += "cull_disabled"
	tmp_code += ";\n"
	
	# Uniforms
	
	tmp_code += color_mode_uniforms[sh_color_mode]
	tmp_code += alpha_uniforms[sh_alpha_mode]
	tmp_code += uv_manip_uniforms[sh_uv_manip_mode]
	tmp_code +="	varying vec2 full_range_uv;\n"
	tmp_code += "	instance uniform float overall_opacity = 1.0;"
	tmp_code += "instance uniform float emission_power = 0.0;"
	if sh_expose_custom:
		tmp_code += "instance uniform vec4 exposed_custom = vec4(1.0);"
	if sh_apply_color_distortion:
		tmp_code += color_distortion_uniform
	
	if sh_flipbook_enabled:
		tmp_code += flipbook_code.uniforms
	# Vertex shader!
	
	tmp_code += "void vertex(){\n"
	tmp_code += """if (!OUTPUT_IS_SRGB) {
		COLOR.rgb = mix(pow((COLOR.rgb + vec3(0.055)) * (1.0 / (1.0 + 0.055)), vec3(2.4)), COLOR.rgb * (1.0 / 12.92), lessThan(COLOR.rgb, vec3(0.04045)));
	}"""
	
	# moon moon
	tmp_code += "vec4 custom_custom = INSTANCE_CUSTOM;\n"
	if sh_expose_custom:
		tmp_code += "custom_custom = exposed_custom;\n"

	# Billboard code
	tmp_code += billboard_codes[sh_billboard_mode]

	# we assign untransformed UV in case we need it for the dissolve
	if sh_uv_dissolve_full_quad:
		tmp_code += "full_range_uv = UV;\n"
	tmp_code += uv_manip_code[sh_uv_manip_mode]
	if sh_uv_swap:
		tmp_code += "UV.xy = UV.yx;\n"
	if not sh_uv_dissolve_full_quad:
		tmp_code += "full_range_uv = UV;\n"
	# flipbook code
	if sh_flipbook_enabled:
		tmp_code += flipbook_code.code
	
	tmp_code += "}"
	
	# Fragment yohooo
	
	tmp_code += "void fragment(){\n"
	
	# first flipbooks sample
	tmp_code += "vec4 color = texture(flipbook, UV);\n"
	tmp_code += "float alpha = COLOR.a;\n"
	tmp_code += "float color_offset = .0;\n"
	
	#ALPHA code
	# we run color code after because we may need alpha information
	tmp_code += alpha_code[sh_alpha_mode]
	if sh_apply_color_distortion:
		tmp_code += color_distortion_code
	tmp_code += color_mode_code[sh_color_mode]
	
	if sh_uv_manip_mode != UvManipulation.NONE:
		tmp_code += """
		if (clamp_alpha_x){
			ALPHA *= step(alpha_clamp_range.x, full_range_uv.x);\n
			ALPHA *= 1.0 - step(alpha_clamp_range.y, full_range_uv.x);\n
		}
		if (clamp_alpha_y){
			ALPHA *= step(alpha_clamp_range.x, full_range_uv.y);\n
			ALPHA *= 1.0 - step(alpha_clamp_range.y, full_range_uv.y);\n
		}
		"""
	tmp_code += "	ALPHA *= overall_opacity;\n"
	tmp_code += "ALBEDO *= 1.0 + emission_power;"
	tmp_code += "}\n"
	shader.code = tmp_code
